@@ -12,7 +12,7 @@ from pony.orm.ormtypes import \
     string_types, numeric_types, comparable_types, SetType, FuncType, MethodType, \
     get_normalized_type_of, normalize_type, coerce_types, are_comparable_types
 from pony.orm import core
-from pony.orm.core import ERDiagramError, EntityMeta, Set, JOIN, AsciiStr
+from pony.orm.core import ERDiagramError, EntityMeta, Set, JOIN, AsciiStr, OptimizationFailed
 
 def check_comparable(left_monad, right_monad, op='=='):
     t1, t2 = left_monad.type, right_monad.type
@@ -118,14 +118,14 @@ class SQLTranslator(ASTTranslator):
             monad.node = node
             if not hasattr(monad, 'aggregated'):
                 for child in node.getChildNodes():
-                    m = getattr(child, 'monad', None) 
+                    m = getattr(child, 'monad', None)
                     if m and getattr(m, 'aggregated', False):
                         monad.aggregated = True
                         break
                 else: monad.aggregated = False
             if not hasattr(monad, 'nogroup'):
                 for child in node.getChildNodes():
-                    m = getattr(child, 'monad', None) 
+                    m = getattr(child, 'monad', None)
                     if m and getattr(m, 'nogroup', False):
                         monad.nogroup = True
                         break
@@ -147,6 +147,7 @@ class SQLTranslator(ASTTranslator):
         translator.left_join = left_join
         translator.optimize = optimize
         translator.from_optimized = False
+        translator.optimization_failed = False
         if not parent_translator: subquery = Subquery(left_join=left_join)
         else: subquery = Subquery(parent_translator.subquery, left_join=left_join)
         translator.subquery = subquery
@@ -158,7 +159,7 @@ class SQLTranslator(ASTTranslator):
         translator.aggregated = False if not optimize else True
         translator.inside_expr = False
         translator.inside_not = False
-        translator.inside_orderby = False
+        translator.inside_order_by = False
         translator.hint_join = False
         translator.aggregated_subquery_paths = set()
         for i, qual in enumerate(tree.quals):
@@ -176,7 +177,7 @@ class SQLTranslator(ASTTranslator):
             if monad:  # Lambda was encountered inside generator
                 assert isinstance(monad, EntityMonad)
                 entity = monad.type.item_type
-                tablerefs[name] = TableRef(subquery, name, entity)               
+                tablerefs[name] = TableRef(subquery, name, entity)
             elif src:
                 iterable = translator.vartypes[src]
                 if not isinstance(iterable, SetType): throw(TranslationError,
@@ -229,7 +230,7 @@ class SQLTranslator(ASTTranslator):
             database = entity._database_
             assert database.schema is not None
             if translator.database is None: translator.database = database
-            elif translator.database is not database: throw(TranslationError, 
+            elif translator.database is not database: throw(TranslationError,
                 'All entities in a query must belong to the same database')
 
             for if_ in qual.ifs:
@@ -345,7 +346,7 @@ class SQLTranslator(ASTTranslator):
                     throw(TranslationError, '%r is valid for numeric attributes only' % aggr_func_name.lower())
                 assert len(translator.expr_columns) == 1
                 column_ast = translator.expr_columns[0]
-            elif aggr_func_name is not 'COUNT': throw(TranslationError, 
+            elif aggr_func_name is not 'COUNT': throw(TranslationError,
                 'Attribute should be specified for %r aggregate function' % aggr_func_name.lower())
             aggr_ast = None
             if aggr_func_name == 'COUNT':
@@ -368,13 +369,13 @@ class SQLTranslator(ASTTranslator):
             sql_ast.append([ 'WHERE' ] + translator.conditions)
 
         if translator.groupby_monads:
-            groupby = [ 'GROUP_BY' ]
-            for m in translator.groupby_monads: groupby.extend(m.getsql())
-            sql_ast.append(groupby)
-        else: groupby = None
+            group_by = [ 'GROUP_BY' ]
+            for m in translator.groupby_monads: group_by.extend(m.getsql())
+            sql_ast.append(group_by)
+        else: group_by = None
 
         if translator.having_conditions:
-            if not groupby: throw(TranslationError,
+            if not group_by: throw(TranslationError,
                 'In order to use aggregated functions such as SUM(), COUNT(), etc., '
                 'query must have grouping columns (i.e. resulting non-aggregated values)')
             sql_ast.append([ 'HAVING' ] + translator.having_conditions)
@@ -487,7 +488,7 @@ class SQLTranslator(ASTTranslator):
         if lambda_expr.varargs: throw(TypeError)
         if lambda_expr.kwargs: throw(TypeError)
         if lambda_expr.defaults: throw(TypeError)
-        iter_name = lambda_expr.argnames[0]            
+        iter_name = lambda_expr.argnames[0]
         cond_expr = lambda_expr.code
         if_expr = ast.GenExprIf(cond_expr)
         name_ast = ast.Name(entity.__name__)
@@ -650,7 +651,7 @@ class JoinedTableRef(object):
             alias = tableref.subquery.get_short_alias(tableref.name_path, right_entity.__name__)
             join_cond = join_tables(m2m_alias, alias, right_m2m_columns, pk_columns)
             tableref.subquery.from_ast.append([ alias, 'TABLE', right_entity._table_, join_cond ])
-        tableref.alias = alias 
+        tableref.alias = alias
         tableref.pk_columns = pk_columns
         tableref.optimized = False
         tableref.joined = True
@@ -706,11 +707,16 @@ class Monad(object):
             expr = [ 'CASE', None, [ [ expr, [ 'VALUE', 1 ] ] ], [ 'VALUE', None ] ]
             count_kind = 'ALL'
         elif len(expr) == 1: expr = expr[0]
-        elif translator.row_value_syntax == True: expr = ['ROW'] + expr
+        elif translator.row_value_syntax == True and translator.dialect != 'Oracle':
+            expr = ['ROW'] + expr
         elif translator.dialect == 'SQLite':
             alias, pk_columns = monad.tableref.make_join(pk_only=False)
             expr = [ 'COLUMN', alias, 'ROWID' ]
-        else: throw(NotImplementedError)
+        else: throw(NotImplementedError,
+                    '%s database provider does not support entities '
+                    'with composite primary keys inside aggregate functions. Got: {EXPR} '
+                    '(you can suggest us how to write SQL for this query)'
+                    % translator.dialect)
         result = translator.ExprMonad.new(translator, int, [ 'COUNT', count_kind, expr ])
         result.aggregated = True
         return result
@@ -731,8 +737,11 @@ class Monad(object):
         expr = monad.getsql()
         if len(expr) == 1: expr = expr[0]
         elif translator.row_value_syntax == True: expr = ['ROW'] + expr
-        else: throw(NotImplementedError, 'Database does not support entities '
-                    'with composite primary keys inside aggregate functions. Got: {EXPR}')
+        else: throw(NotImplementedError,
+                    '%s database provider does not support entities '
+                    'with composite primary keys inside aggregate functions. Got: {EXPR} '
+                    '(you can suggest us how to write SQL for this query)'
+                    % translator.dialect)
         if func_name == 'AVG': result_type = float
         else: result_type = expr_type
         result = translator.ExprMonad.new(translator, result_type, [ func_name, expr ])
@@ -798,16 +807,6 @@ class MethodMonad(Monad):
     def __neg__(monad): raise_forgot_parentheses(monad)
     def abs(monad): raise_forgot_parentheses(monad)
 
-class DatabaseMonad(Monad):
-    def __init__(monad, translator, database):
-        Monad.__init__(monad, translator, 'DATABASE')
-        Monad.database = database
-    def getattr(monad, attrname):
-        database = monad.database
-        entity = getattr(database, attrname)
-        if not isinstance(entity, EntityMeta): throw(NotImplementedError)
-        return EntityMonad(monad.translator, entity)
-
 class EntityMonad(Monad):
     def __init__(monad, translator, entity):
         Monad.__init__(monad, translator, SetType(entity))
@@ -817,8 +816,6 @@ class EntityMonad(Monad):
             throw(TranslationError, 'All entities in a query must belong to the same database')
     def __getitem__(monad, *args):
         throw(NotImplementedError)
-    def call_query(monad):
-        throw(NotImplementedError)
 
 class ListMonad(Monad):
     def __init__(monad, translator, items):
@@ -901,7 +898,7 @@ class DateMixin(MonadMixin):
     attr_year = datetime_attr_factory('YEAR')
     attr_month = datetime_attr_factory('MONTH')
     attr_day = datetime_attr_factory('DAY')
-    
+
 class DatetimeMixin(DateMixin):
     def mixin_init(monad):
         assert monad.type is datetime
@@ -956,7 +953,7 @@ class StringMixin(MonadMixin):
             expr_sql = monad.getsql()[0]
 
             if start is None: start = translator.ConstMonad.new(translator, 0)
-            
+
             if isinstance(start, translator.NumericConstMonad):
                 if start.value < 0: throw(NotImplementedError, 'Negative indices are not supported in string slice {EXPR}')
                 start_sql = [ 'VALUE', start.value + 1 ]
@@ -981,10 +978,10 @@ class StringMixin(MonadMixin):
 
             sql = [ 'SUBSTR', expr_sql, start_sql, len_sql ]
             return translator.StringExprMonad(translator, monad.type, sql)
-        
+
         if isinstance(monad, translator.StringConstMonad) and isinstance(index, translator.NumericConstMonad):
             return translator.ConstMonad.new(translator, monad.value[index.value])
-        if index.type is not int: throw(TypeError, 
+        if index.type is not int: throw(TypeError,
             'String indices must be integers. Got %r in expression {EXPR}' % type2str(index.type))
         expr_sql = monad.getsql()[0]
         if isinstance(index, translator.NumericConstMonad):
@@ -1061,7 +1058,7 @@ class StringMixin(MonadMixin):
         return monad.strip(chars, 'LTRIM')
     def call_rstrip(monad, chars=None):
         return monad.strip(chars, 'RTRIM')
-    
+
 class ObjectMixin(MonadMixin):
     def mixin_init(monad):
         assert isinstance(monad.type, EntityMeta)
@@ -1123,7 +1120,7 @@ class AttrMonad(Monad):
             offset = attr.pk_columns_offset
             columns = parent_columns[offset:offset+len(attr.columns)]
         return [ [ 'COLUMN', alias, column ] for column in columns ]
-        
+
 class ObjectAttrMonad(ObjectMixin, AttrMonad):
     def __init__(monad, parent, attr):
         AttrMonad.__init__(monad, parent, attr)
@@ -1242,7 +1239,7 @@ class BufferConstMonad(BufferMixin, ConstMonad): pass
 class StringConstMonad(StringMixin, ConstMonad):
     def len(monad):
         return monad.translator.ConstMonad.new(monad.translator, len(monad.value))
-    
+
 class NumericConstMonad(NumericMixin, ConstMonad): pass
 class DateConstMonad(DateMixin, ConstMonad): pass
 class DatetimeConstMonad(DatetimeMixin, ConstMonad): pass
@@ -1275,7 +1272,7 @@ class BoolExprMonad(BoolMonad):
         else: return translator.NotMonad(translator, sql)
         return translator.BoolExprMonad(translator, negated_sql)
 
-cmp_ops = { '>=' : 'GE', '>' : 'GT', '<=' : 'LE', '<' : 'LT' }        
+cmp_ops = { '>=' : 'GE', '>' : 'GT', '<=' : 'LE', '<' : 'LT' }
 
 cmp_negate = { '<' : '>=', '<=' : '>', '==' : '!=', 'is' : 'is not' }
 cmp_negate.update((b, a) for a, b in cmp_negate.items())
@@ -1398,7 +1395,7 @@ class FuncDateMonad(FuncMonad):
     def call(monad, year, month, day):
         translator = monad.translator
         for x, name in zip((year, month, day), ('year', 'month', 'day')):
-            if not isinstance(x, translator.NumericMixin) or x.type is not int: throw(TypeError, 
+            if not isinstance(x, translator.NumericMixin) or x.type is not int: throw(TypeError,
                 "'%s' argument of date(year, month, day) function must be of 'int' type. Got: %r" % (name, type2str(x.type)))
             if not isinstance(x, translator.ConstMonad): throw(NotImplementedError)
         return translator.ConstMonad.new(translator, date(year.value, month.value, day.value))
@@ -1411,7 +1408,7 @@ class FuncDatetimeMonad(FuncDateMonad):
     def call(monad, *args):
         translator = monad.translator
         for x, name in zip(args, ('year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond')):
-            if not isinstance(x, translator.NumericMixin) or x.type is not int: throw(TypeError, 
+            if not isinstance(x, translator.NumericMixin) or x.type is not int: throw(TypeError,
                 "'%s' argument of datetime(...) function must be of 'int' type. Got: %r" % (name, type2str(x.type)))
             if not isinstance(x, translator.ConstMonad): throw(NotImplementedError)
         return translator.ConstMonad.new(translator, datetime(*tuple(arg.value for arg in args)))
@@ -1468,7 +1465,7 @@ def minmax(monad, sqlop, *args):
     translator = monad.translator
     t = args[0].type
     if t == 'METHOD': raise_forgot_parentheses(args[0])
-    if t not in comparable_types: throw(TypeError, 
+    if t not in comparable_types: throw(TypeError,
         "Value of type %r is not valid as argument of %r function in expression {EXPR}"
         % (type2str(t), sqlop.lower()))
     for arg in args[1:]:
@@ -1484,14 +1481,14 @@ class FuncSelectMonad(FuncMonad):
     func = core.select
     def call(monad, queryset):
         translator = monad.translator
-        if not isinstance(queryset, translator.QuerySetMonad): throw(TypeError, 
+        if not isinstance(queryset, translator.QuerySetMonad): throw(TypeError,
             "'select' function expects generator expression, got: {EXPR}")
         return queryset
 
 class FuncExistsMonad(FuncMonad):
     func = core.exists
     def call(monad, arg):
-        if not isinstance(arg, monad.translator.SetMixin): throw(TypeError, 
+        if not isinstance(arg, monad.translator.SetMixin): throw(TypeError,
             "'exists' function expects generator expression or collection, got: {EXPR}")
         return arg.nonzero()
 
@@ -1564,9 +1561,29 @@ class AttrSetMonad(SetMixin, Monad):
                 return translator.BoolExprMonad(translator, [ sqlop, [ 'ROW' ] + item.getsql(), subquery_ast ])
             else:
                 conditions += [ [ 'EQ', expr1, expr2 ] for expr1, expr2 in izip(item.getsql(), expr_list) ]
-                subquery_ast = [ not_in and 'NOT_EXISTS' or 'EXISTS', from_ast, [ 'WHERE' ] + conditions ] 
+                subquery_ast = [ not_in and 'NOT_EXISTS' or 'EXISTS', from_ast, [ 'WHERE' ] + conditions ]
                 return translator.BoolExprMonad(translator, subquery_ast)
-        else: throw(NotImplementedError)
+        elif not not_in:
+            translator.distinct = True
+            tableref = monad.make_tableref(translator.subquery)
+            alias, columns = monad.tableref.make_join(pk_only=True)
+            expr_ast = sqland([ [ 'EQ', [ 'COLUMN', alias, column ], expr ]  for column, expr in zip(columns, item.getsql()) ])
+            return translator.BoolExprMonad(translator, expr_ast)
+        else:
+            subquery = Subquery(translator.subquery)
+            tableref = monad.make_tableref(subquery)
+            attr = monad.attr
+            alias, columns = tableref.make_join(pk_only=attr.reverse)
+            subquery.expr_list = monad.make_expr_list()
+            if not attr.reverse: columns = attr.columns
+            from_ast = translator.subquery.from_ast
+            from_ast[0] = 'LEFT_JOIN'
+            from_ast.extend(subquery.from_ast[1:])
+            conditions = [ [ 'EQ', [ 'COLUMN', alias, column ], expr ]  for column, expr in zip(columns, item.getsql()) ]
+            conditions.extend(subquery.conditions)
+            from_ast[-1][-1] = sqland([ from_ast[-1][-1] ] + conditions)
+            expr_ast = sqland([ [ 'IS_NULL', [ 'COLUMN', alias, column ] ] for column in columns ])
+            return translator.BoolExprMonad(translator, expr_ast)
     def getattr(monad, name):
         try: return Monad.getattr(monad, name)
         except AttributeError: pass
@@ -1601,6 +1618,7 @@ class AttrSetMonad(SetMixin, Monad):
         elif len(expr_list) == 1:
             make_aggr = lambda expr_list: [ 'COUNT', 'DISTINCT' ] + expr_list
         elif translator.dialect == 'Oracle':
+            if monad.tableref.name_path == translator.optimize: raise OptimizationFailed
             extra_grouping = True
             if translator.hint_join: make_aggr = lambda expr_list: [ 'COUNT', 'ALL' ]
             else: make_aggr = lambda expr_list: [ 'COUNT', 'ALL', [ 'COUNT', 'ALL' ] ]
@@ -1637,16 +1655,16 @@ class AttrSetMonad(SetMixin, Monad):
         translator = monad.translator
         item_type = monad.type.item_type
 
-        if func_name in ('SUM', 'AVG'):        
-            if item_type not in numeric_types: throw(TypeError, 
+        if func_name in ('SUM', 'AVG'):
+            if item_type not in numeric_types: throw(TypeError,
                 "Function %s() expects query or items of numeric type, got %r in {EXPR}"
                 % (func_name.lower(), type2str(item_type)))
         elif func_name in ('MIN', 'MAX'):
-            if item_type not in comparable_types: throw(TypeError, 
+            if item_type not in comparable_types: throw(TypeError,
                 "Function %s() expects query or items of comparable type, got %r in {EXPR}"
                 % (func_name.lower(), type2str(item_type)))
         else: assert False
-            
+
         if monad.forced_distinct and func_name in ('SUM', 'AVG'):
             make_aggr = lambda expr_list: [ func_name ] + expr_list + [ True ]
         else:
@@ -1682,7 +1700,7 @@ class AttrSetMonad(SetMixin, Monad):
         if isinstance(parent, ObjectMixin): parent_tableref = parent.tableref
         elif isinstance(parent, translator.AttrSetMonad): parent_tableref = parent.make_tableref(subquery)
         else: assert False
-        if attr.reverse: 
+        if attr.reverse:
             name_path = parent_tableref.name_path + '-' + attr.name
             monad.tableref = subquery.get_tableref(name_path) \
                              or subquery.add_tableref(name_path, parent_tableref, attr)
@@ -1723,7 +1741,7 @@ class AttrSetMonad(SetMixin, Monad):
         from_ast = subquery.from_ast
         inner_conditions = subquery.conditions
         outer_conditions = subquery.outer_conditions
-        
+
         groupby_columns = [ inner_column[:] for cond, outer_column, inner_column in outer_conditions ]
         assert len(set(alias for _, alias, column in groupby_columns)) == 1
 
@@ -1756,7 +1774,7 @@ class AttrSetMonad(SetMixin, Monad):
                 tname, cname = inner_column[1:]
                 new_name = col_mapping[tname, cname]
                 outer_conditions[i] = [ cond, outer_column, [ 'COLUMN', inner_alias, new_name ] ]
-                
+
         subquery_columns = [ 'ALL' ]
         for column_ast in groupby_columns:
             assert column_ast[0] == 'COLUMN'
@@ -1772,11 +1790,10 @@ class AttrSetMonad(SetMixin, Monad):
         for cond in outer_conditions: cond[2][1] = alias
         translator.subquery.from_ast.append([ alias, 'SELECT', subquery_ast, sqland(outer_conditions) ])
         expr_ast = [ 'COLUMN', alias, expr_name ]
-        if coalesce_to_zero: expr_ast = [ 'COALESCE', expr_ast, [ 'VALUE', 0 ] ]        
+        if coalesce_to_zero: expr_ast = [ 'COALESCE', expr_ast, [ 'VALUE', 0 ] ]
         return expr_ast, False
     def _subselect(monad):
         if monad.subquery is not None: return monad.subquery
-        parent = monad.parent
         attr = monad.attr
         translator = monad.translator
         subquery = Subquery(translator.subquery)
@@ -1789,7 +1806,6 @@ class AttrSetMonad(SetMixin, Monad):
         monad.subquery = subquery
         return subquery
     def getsql(monad, subquery=None):
-        parent = monad.parent
         if subquery is None: subquery = monad.translator.subquery
         monad.make_tableref(subquery)
         return monad.make_expr_list()
@@ -1865,28 +1881,68 @@ class QuerySetMonad(SetMixin, Monad):
         sub = monad.subtranslator
         columns_ast = sub.expr_columns
         conditions = sub.conditions[:]
-        if not translator.hint_join:
-            if len(columns_ast) == 1 or translator.row_value_syntax:
-                select_ast = [ 'ALL' ] + columns_ast
-                subquery_ast = [ 'SELECT', select_ast, sub.subquery.from_ast ]
-                subquery_expr = sub.tree.expr.monad
-                if isinstance(subquery_expr, translator.AttrMonad) and not subquery_expr.attr.nullable: pass
-                else: conditions += [ [ 'IS_NOT_NULL', column_ast ] for column_ast in sub.expr_columns ]
-                if conditions: subquery_ast.append([ 'WHERE' ] + conditions)
-                if len(columns_ast) == 1: expr_ast = item.getsql()[0]
-                else: expr_ast = [ 'ROW' ] + item.getsql()
-                sql_ast = [ not_in and 'NOT_IN' or 'IN', expr_ast, subquery_ast ]
-                return translator.BoolExprMonad(translator, sql_ast)
+        if isinstance(item, translator.ListMonad):
+            item_columns = []
+            for subitem in item.items: item_columns.extend(subitem.getsql())
+        else: item_columns = item.getsql()
+        if translator.hint_join:
+            subquery = translator.subquery
+            if not not_in:
+                translator.distinct = True
+                if subquery.from_ast[0] == 'FROM':
+                    subquery.from_ast[0] = 'INNER_JOIN'
             else:
-                if isinstance(item, translator.ListMonad):
-                    item_columns = []
-                    for subitem in item.items: item_columns.extend(subitem.getsql())
-                else: item_columns = item.getsql()
-                conditions += [ [ 'EQ', expr1, expr2 ] for expr1, expr2 in izip(item_columns, columns_ast) ]
-                subquery_ast = [ not_in and 'NOT_EXISTS' or 'EXISTS', sub.subquery.from_ast, [ 'WHERE' ] + conditions ]
-                return translator.BoolExprMonad(translator, subquery_ast)
-        else: throw(NotImplementedError)
-    def nonzero(monad):        
+                subquery.left_join = True
+                subquery.from_ast[0] = 'LEFT_JOIN'
+            col_names = set()
+            next = subquery.expr_counter.next
+            new_names = []
+            exprs = []
+            for column_ast in columns_ast:
+                if column_ast[0] == 'COLUMN':
+                    tab_name, col_name = column_ast[1:]
+                    if col_name not in col_names:
+                        col_names.add(col_name)
+                        new_names.append(col_name)
+                        exprs.append([ 'AS', column_ast, col_name ])
+                        continue
+                new_name = 'expr-%d' % next()
+                new_names.append(new_name)
+                exprs.append([ 'AS', column_ast, new_name ])
+            from_ast = sub.subquery.from_ast[:]
+            if len(from_ast[1]) == 4:
+                outer_conditions = from_ast[1][-1]
+                from_ast[1] = from_ast[:-1]
+            else: outer_conditions = []
+            subquery_ast = [ [ 'ALL' ] + exprs, from_ast ]
+            subquery_expr = sub.tree.expr.monad
+            if isinstance(subquery_expr, translator.AttrMonad) and not subquery_expr.attr.nullable: pass
+            else: conditions += [ [ 'IS_NOT_NULL', column_ast ] for column_ast in sub.expr_columns ]
+            if conditions: subquery_ast.append([ 'WHERE' ] + conditions)
+            alias = subquery.get_short_alias(None, 't')
+            outer_conditions.extend([ [ 'EQ', item_column, [ 'COLUMN', alias, new_name ] ]
+                                    for item_column, new_name in izip(item_columns, new_names) ])
+            subquery.from_ast.append([ alias, 'SELECT', subquery_ast, sqland(outer_conditions) ])
+            if not_in: result_expr = sqland([ [ 'IS_NULL', [ 'COLUMN', alias, new_name ] ]
+                                              for new_name in new_names ])
+            else: result_expr = [ 'EQ', [ 'VALUE', 1 ], [ 'VALUE', 1 ] ]
+            return translator.BoolExprMonad(translator, result_expr)
+        if len(columns_ast) == 1 or translator.row_value_syntax:
+            select_ast = [ 'ALL' ] + columns_ast
+            subquery_ast = [ 'SELECT', select_ast, sub.subquery.from_ast ]
+            subquery_expr = sub.tree.expr.monad
+            if isinstance(subquery_expr, translator.AttrMonad) and not subquery_expr.attr.nullable: pass
+            else: conditions += [ [ 'IS_NOT_NULL', column_ast ] for column_ast in sub.expr_columns ]
+            if conditions: subquery_ast.append([ 'WHERE' ] + conditions)
+            if len(columns_ast) == 1: expr_ast = item_columns[0]
+            else: expr_ast = [ 'ROW' ] + item_columns
+            sql_ast = [ not_in and 'NOT_IN' or 'IN', expr_ast, subquery_ast ]
+            return translator.BoolExprMonad(translator, sql_ast)
+        else:
+            conditions += [ [ 'EQ', expr1, expr2 ] for expr1, expr2 in izip(item_columns, columns_ast) ]
+            subquery_ast = [ not_in and 'NOT_EXISTS' or 'EXISTS', sub.subquery.from_ast, [ 'WHERE' ] + conditions ]
+            return translator.BoolExprMonad(translator, subquery_ast)
+    def nonzero(monad):
         sub = monad.subtranslator
         sql_ast = [ 'EXISTS', sub.subquery.from_ast, [ 'WHERE' ] + sub.conditions ]
         translator = monad.translator
@@ -1945,14 +2001,14 @@ class QuerySetMonad(SetMixin, Monad):
         sub = monad.subtranslator
         expr_type = sub.expr_type
         if func_name in ('SUM', 'AVG'):
-            if expr_type not in numeric_types: throw(TypeError, 
+            if expr_type not in numeric_types: throw(TypeError,
                 "Function %s() expects query or items of numeric type, got %r in {EXPR}"
                 % (func_name.lower(), type2str(expr_type)))
         elif func_name in ('MIN', 'MAX'):
-            if expr_type not in comparable_types: throw(TypeError, 
+            if expr_type not in comparable_types: throw(TypeError,
                 "Function %s() cannot be applied to type %r in {EXPR}"
                 % (func_name.lower(), type2str(expr_type)))
-        else: assert False        
+        else: assert False
         assert len(sub.expr_columns) == 1
         select_ast = [ 'AGGREGATES', [ func_name, sub.expr_columns[0] ] ]
         result_type = func_name == 'AVG' and float or expr_type
